#! /usr/bin/env python
# coding: utf-8

import sys
import logging
import argparse
from JYTools.JYWorker import RedisStat, RedisQueue

__author__ = '鹛桑够'

# logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
logging.basicConfig()

arg_man = argparse.ArgumentParser()
arg_man.add_argument("--debug", dest="debug", help="debug mode, print debug msg", action="store_true", default=False)


def parse_args():
    args = arg_man.parse_args()
    if args.debug is True:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(levelname)s %(message)s")
    return args


def empty_help():
    if len(sys.argv) <= 1:
        sys.argv.append("-h")


def list_queue():
    arg_man.add_argument("-w", "--work-tag", dest="work_tag", help="work tag", metavar="")
    args = parse_args()
    rs = RedisStat()
    if args.work_tag is None:
        qd = rs.list_queue()
        for key in qd:
            print(key)
    else:
        lqd = rs.list_queue_detail(args.work_tag)
        for item in lqd:
            print(item)


def list_worry_queue():
    rs = RedisStat()
    wq = rs.list_worry_queue()
    for item in wq:
        print(item)


def list_heartbeat():
    arg_man.add_argument("-w", "--work-tag", dest="work_tag", help="work tag", metavar="")
    rs = RedisStat()
    args = parse_args()
    if args.work_tag is None:
        ws = rs.list_heartbeat()
        for item in ws:
            print(item)
    else:
        hd = rs.list_heartbeat_detail(args.work_tag)
        print(hd)


def delete_heartbeat():
    empty_help()
    rs = RedisStat()
    arg_man.add_argument("-w", "--work-tag", dest="work_tag", help="work tag", metavar="", nargs="*", default=[])
    args = parse_args()
    for item in args.work_tag:
        print("delete heartbeat %s" % item)
        rs.delete_heartbeat(item)


def list_worker():
    arg_man.add_argument("-w", "--work-tag", dest="work_tag", help="work tag", metavar="")
    rs = RedisStat()
    args = parse_args()
    if args.work_tag is None:
        ws = rs.list_worker()
        for item in ws:
            print(item)
    else:
        hd = rs.list_worker_detail(args.work_tag)
        print(hd)


def wash_worker():
    empty_help()
    arg_man.add_argument("-w", "--work-tag", dest="work_tag", help="work tag", metavar="", required=True)
    arg_man.add_argument("-n", "--num", dest="num", help="num of wash package to send", metavar="", type=int, default=1)
    args = parse_args()
    r_queue = RedisQueue()
    r_queue.wash_worker(args.work_tag, args.num)


if __name__ == "__main__":
    sys.argv.append("--debug")
    sys.argv.extend(["-w", "JYAnalysisDAG", "APP7"])
    # wash_worker()
    logging.info("ssd")
    delete_heartbeat()


#! /usr/bin/env python
# coding: utf-8

import re
from time import time
from JYTools.StringTool import is_string, join_decode
from ._Task import TaskStatus
from ._redis import RedisWorker

__author__ = 'meisanggou'


class DAGWorker(RedisWorker):
    expect_params_type = dict
    ref_compile = re.compile(r"^(\d{1,10})((&\d+|&*[a-z])\w{0,60})(\**)$", re.I)

    def __init__(self, conf_path=None, heartbeat_value=None, is_brother=False, work_tag=None, log_dir=None,
                 redis_host=None, redis_password=None, redis_port=None, redis_db=None, section_name="Redis", **kwargs):
        self.agent_tag = kwargs.pop("agent_tag", None)
        RedisWorker.__init__(self, conf_path, heartbeat_value, is_brother, work_tag, log_dir, redis_host,
                             redis_password, redis_port, redis_db, section_name, **kwargs)

    def push_task(self, key, params, work_tag=None, sub_key=None, report_tag=None, is_report=False):
        if self.agent_tag is not None:
            if work_tag is not None and work_tag not in (self.work_tag, self.upload_log_tag) and is_report is False:
                params = dict(work_tag=work_tag, params=params)
                work_tag = self.agent_tag
        self._push_task(key, params, work_tag, sub_key, report_tag, is_report=is_report)

    @staticmethod
    def split_ref(ref_str):
        """

        :param ref_str: index+字母开头的key index+&+数字开头的key index+&+字母开头的key
                        最后可以加入*结尾，也可以不加入，暂时定义为required
        :return:
        """
        if ref_str[0] == "&":
            ref_str = ref_str[1:]
        match_r = DAGWorker.ref_compile.match(ref_str)
        if match_r is None:
            return None
        ref_index = int(match_r.groups()[0])
        ref_key = match_r.groups()[1]
        required = match_r.groups()[2]
        if ref_key[0] == "&":
            ref_key = ref_key[1:]
        if required == "*":
            required = False
        else:
            required = True
        return dict(index=ref_index, key=ref_key, required=required)

    @staticmethod
    def exist_loop(params):
        tl = params["task_list"]
        assert isinstance(tl, list)
        task_len = len(tl)
        assert task_len > 0
        rs_l = [dict(quotes=list(), next=list(), index=i) for i in range(task_len)]
        for index in range(task_len):
            task_item = tl[index]
            assert isinstance(task_item, dict)
            for k, v in task_item.items():
                if k.startswith("input_") is False:
                    continue
                if is_string(v) is False:
                    continue
                ref_d = DAGWorker.split_ref(v)
                if ref_d is None:
                    continue
                if ref_d["index"] < 0 or ref_d["index"] > task_len:
                    raise ValueError("")  # out of index
                if ref_d["index"] not in rs_l[index]["quotes"]:
                    rs_l[index]["quotes"].append(ref_d["index"])
                    if ref_d["index"] > 0:
                        rs_l[ref_d["index"] - 1]["next"].append(index)
        completed_queue = [0]

        while True:
            completed_num = 0
            for index in range(task_len):
                if index + 1 in completed_queue:
                    continue
                rs_item = rs_l[index]
                q_len = len(rs_item["quotes"])
                for i in range(q_len - 1, -1, -1):
                    if rs_item["quotes"][i] in completed_queue:
                        rs_item["quotes"].remove(rs_item["quotes"][i])
                if len(rs_item["quotes"]) <= 0:
                    completed_queue.append(index + 1)
                    completed_num += 1
                    continue
            if len(completed_queue) == task_len + 1:
                return False
            if completed_num == 0:
                return True
        return False

    @staticmethod
    def find_loop2(params):
        tl = params["task_list"]
        assert isinstance(tl, list)
        task_len = len(tl)
        assert task_len > 0
        rs_l = [dict(quotes=list(), next=list(), index=i) for i in range(task_len)]
        for index in range(task_len):
            task_item = tl[index]
            assert isinstance(task_item, dict)
            for k, v in task_item.items():
                if k.startswith("input_") is False:
                    continue
                if is_string(v) is False:
                    continue
                ref_d = DAGWorker.split_ref(v)
                if ref_d is None:
                    continue
                if ref_d["index"] < 0 or ref_d["index"] > task_len:
                    raise ValueError("")  # out of index
                if ref_d["index"] not in rs_l[index]["quotes"]:
                    rs_l[index]["quotes"].append(ref_d["index"])
                    if ref_d["index"] > 0:
                        rs_l[ref_d["index"] - 1]["next"].append(index)

        for index in range(task_len):
            def link(j, l):
                if j + 1 in l:
                    l.append(j + 1)
                    return l
                if len(rs_l[j]["next"]) <= 0:
                    return None
                l.append(j + 1)
                for n_item in rs_l[j]["next"]:
                    lr_l = link(n_item, l)
                    if lr_l is not None:
                        return lr_l
                l.remove(l[-1])
                return None
            r_l = link(index, list())
            if r_l is not None:
                return r_l

        return None

    @staticmethod
    def find_loop(params):
        tl = params["task_list"]
        assert isinstance(tl, list)
        task_len = len(tl)
        assert task_len > 0
        rs_l = [dict(quotes=list(), next=list(), index=i) for i in range(task_len + 1)]
        for index in range(task_len):
            task_item = tl[index]
            assert isinstance(task_item, dict)
            for k, v in task_item.items():
                if k.startswith("input_") is False:
                    continue
                if is_string(v) is False:
                    continue
                ref_d = DAGWorker.split_ref(v)
                if ref_d is None:
                    continue
                if ref_d["index"] < 0 or ref_d["index"] > task_len:
                    raise ValueError("")  # out of index
                if ref_d["index"] not in rs_l[index + 1]["quotes"]:
                    rs_l[index + 1]["quotes"].append(ref_d["index"])
                    rs_l[ref_d["index"]]["next"].append(index + 1)

        for index in range(1, task_len):
            def link(j, l):
                if len(rs_l[j]["next"]) <= 0:
                    return None
                for n_item in rs_l[j]["next"]:
                    if n_item in l:
                        l.append(n_item)
                        return l[l.index(n_item):]
                    l.append(n_item)
                    lr_l = link(n_item, l)
                    if lr_l is not None:
                        return lr_l
                    l.remove(l[-1])
                return None
            r_l = link(index, [index])
            if r_l is not None:
                return r_l

        return None

    @staticmethod
    def find_loop3(params):
        tl = params["task_list"]
        assert isinstance(tl, list)
        task_len = len(tl)
        assert task_len > 0
        rs_l = [dict(quotes=list(), next=list(), index=i) for i in range(task_len + 1)]
        for index in range(task_len):
            task_item = tl[index]
            assert isinstance(task_item, dict)
            for k, v in task_item.items():
                if k.startswith("input_") is False:
                    continue
                if is_string(v) is False:
                    continue
                ref_d = DAGWorker.split_ref(v)
                if ref_d is None:
                    continue
                if ref_d["index"] < 0 or ref_d["index"] > task_len:
                    raise ValueError("")  # out of index
                if ref_d["index"] not in rs_l[index + 1]["quotes"]:
                    rs_l[index + 1]["quotes"].append(ref_d["index"])
                    rs_l[ref_d["index"]]["next"].append(index + 1)
        passed_point = [False] * (task_len + 1)
        for index in range(1, task_len):
            def link(j, l):
                if passed_point[j] is not False:
                    return None
                if len(rs_l[j]["next"]) <= 0:
                    passed_point[j] = True
                    return None
                for n_item in rs_l[j]["next"]:
                    if n_item in l:
                        l.append(n_item)
                        return l[l.index(n_item):]
                    l.append(n_item)
                    lr_l = link(n_item, l)
                    if lr_l is not None:
                        return lr_l
                    l.remove(l[-1])
                passed_point[j] = True
                return None
            r_l = link(index, [index])
            if r_l is not None:
                return r_l

        return None

    def get_sub_task_sub_key(self, index=None, task_index=None):
        if index is not None:
            task_index = index + 1
        if self.current_task.task_sub_key is None:
            sub_key = "%s" % task_index
        else:
            sub_key = "%s_%s" % (self.current_task.task_sub_key, task_index)
        return sub_key

    def handle_report_task(self):
        r_task = self.current_task.task_params
        sp_keys = self.current_task.task_sub_key.rsplit("_", 1)
        if len(sp_keys) == 2:
            self.current_task.task_sub_key = sp_keys[0]
        else:
            self.current_task.task_sub_key = None
        reporter_sub_key = int(sp_keys[-1])  # 子任务在父任务中的位置 位置从1开始
        self.task_log("Task ", reporter_sub_key, " Report")
        task_status = r_task.task_status
        task_message = r_task.task_message
        self.task_log("Task ", reporter_sub_key, " Status Is ", task_status)
        self.set_task_item(reporter_sub_key, "task_status", task_status)
        self.set_task_item(reporter_sub_key, "task_message", task_message)
        if r_task["sub_task_detail"] is not None:
            self.set_task_item(reporter_sub_key, "task_list", r_task.sub_task_detail)
        if task_status != TaskStatus.SUCCESS:
            self.set_task_item(0, "task_message", task_message)
            self.set_task_item(0, "task_fail_index", reporter_sub_key)
            self.task_log("Sub Task ", r_task.task_sub_key, " ", r_task.work_tag, " Failed", level="ERROR")
            self.fail_pipeline("Sub Task ", r_task.task_sub_key, " ", r_task.work_tag, " Failed")
        if isinstance(r_task.task_output, dict):
            for output_key in r_task.task_output.keys():
                self.set_task_item(reporter_sub_key, "output_%s" % output_key, r_task.task_output[output_key])
        self.set_task_item(reporter_sub_key, "start_time", r_task.start_time)
        self.set_task_item(reporter_sub_key, "end_time", r_task.end_time)
        self.set_task_item(reporter_sub_key, "finished_time", time())
        # 获取当前pipeline状态
        pipeline_status = self.get_task_item(0, hash_key="task_status")
        if pipeline_status == TaskStatus.FAIL:
            return self.try_finish_pipeline()
        self.handle_task(self.current_task.task_key, None)

    def format_pipeline(self, key, params):
        if "task_list" not in params:
            self.set_current_task_invalid("Need task_list")
        if "name" in params:
            self.set_task_item(0, "task_name", params["name"])
        task_list = params["task_list"]
        if isinstance(task_list, list) is False:
            self.set_current_task_invalid("Need tuple task_list. Now Is ", type(task_list))
        task_len = len(task_list)
        if task_len <= 0:
            self.set_current_task_invalid("At Least One Task")
        for index in range(task_len):
            self.del_task_item(index + 1)
            task_item = task_list[index]
            if isinstance(task_item, dict) is False:
                self.set_current_task_invalid("Task ", index + 1, " Desc Not Dict")
            task_type = task_item.get("task_type", "app")
            if task_type not in ("task", "pipeline", "app", "repeat-app", "repeat-pipeline"):
                self.set_current_task_invalid("Task ", index + 1, " Invalid Task Type ", task_type)
            task_item["task_type"] = task_type
            if task_type.endswith("pipeline"):
                task_item["work_tag"] = self.work_tag
            if "work_tag" not in task_item:
                self.set_current_task_invalid("Task ", index + 1, " work_tag Not Found")
            if "task_status" in task_item:
                if task_item["task_status"] != TaskStatus.SUCCESS:
                    del task_item["task_status"]
            if "task_output" in task_item:
                if isinstance(task_item["task_output"], dict):
                    for key in task_item["task_output"].keys():
                        if "output_%s" % key not in task_item:
                            task_item["output_%s" % key] = task_item["task_output"][key]
        task_output = params.get("task_output", dict())
        for key in params:
            if key.startswith("input_"):
                self.set_task_item(0, key, params[key])
            elif key.startswith("output_"):
                task_output[key[7:]] = params[key]
        self.set_task_item(0, "task_len", task_len)
        if self.current_task.task_report_tag is not None:
            self.set_task_item(0, "report_tag", self.current_task.task_report_tag)
            self.current_task.task_report_tag = None  # 真正执行完后才进行report
        self.set_task_item(0, "task_output", task_output)
        self.set_task_item(0, "start_time", time())
        for index in range(task_len):
            task_item = task_list[index]
            for key in task_item.keys():
                self.set_task_item(index + 1, key, task_item[key])

    def completed_pipeline(self):
        task_len = self.get_task_item(0, hash_key="task_len")
        task_output = self.get_task_item(0, hash_key="task_output")
        if task_output is not None:
            outputs = dict()
            for out_key in task_output.keys():
                out_value = task_output[out_key]
                if is_string(out_value) is True and out_value.startswith("&"):
                    ref_r, ref_info = self.analysis_ref(out_value[1:], None, task_len)
                    if ref_r is False:
                        self.task_log(ref_info, level="WARNING")
                        continue
                    if ref_info is None:
                        continue
                    out_value = ref_info["ref_output"]
                    outputs[out_key] = out_value
                    self.set_task_item(0, "output_%s" % out_key, out_value)
                elif isinstance(out_value, list):
                    for sub_i in range(len(out_value)):
                        sub_v = out_value[sub_i]
                        if is_string(sub_v) is False or sub_v.startswith("&") is False:
                            continue
                        ref_r, ref_info = self.analysis_ref(sub_v[1:], None, task_len)
                        if ref_r is False:
                            self.task_log(ref_info, level="WARNING")
                            continue
                        if ref_info is None:
                            continue
                        out_value[sub_i] = ref_info["ref_output"]
                    outputs[out_key] = out_value
            self.set_multi_output(**outputs)
        self.package_task_item(task_len)
        pipeline_report_tag = self.get_task_item(0, hash_key="report_tag")
        if pipeline_report_tag is not None:
            self.current_task.is_report_task = False
            self.current_task.task_report_tag = pipeline_report_tag
        self.clear_task_item(task_len)

    def try_remove_running_task(self):
        self.task_log("Try to remove or stop running task")
        task_len = self.get_task_item(0, hash_key="task_len")
        if task_len is None:
            self.set_current_task_error("Not Found Pipeline Task Len")
            return False
        for index in range(task_len):
            task_index = index + 1
            if self.get_task_item(task_index, "task_status") != TaskStatus.RUNNING:
                continue
            work_tag = self.get_task_item(task_index, "work_tag")
            self.task_log("Try to remove task %s %s" % (task_index, work_tag))
            sub_key = self.get_sub_task_sub_key(task_index=task_index)
            count = self.stat_man.remove_queue_task(work_tag, self.current_task.task_key, self.work_tag, sub_key)
            if count > 0:
                self.task_log("Remove task %s %s SUCCESS" % (task_index, work_tag))
                self.set_task_item(task_index, hash_key="task_status", hash_value=TaskStatus.NONE)

    def try_finish_pipeline(self):
        """
        若无正在运行的任务，清理pipeline的调度信息，打包运行结果，汇报结果。返回True
        若有正在运行的任务。返回False
        :return:
        """
        task_len = self.get_task_item(0, hash_key="task_len")
        if task_len is None:
            self.set_current_task_error("Not Found Pipeline Task Len")
            return False
        running_count = 0
        for index in range(task_len):
            if self.get_task_item(index + 1, "task_status") == TaskStatus.RUNNING:
                running_count += 1
        if running_count != 0:
            return False
        self.package_task_item(task_len)
        pipeline_report_tag = self.get_task_item(0, hash_key="report_tag")
        if pipeline_report_tag is not None:
            self.current_task.is_report_task = False
            self.current_task.task_report_tag = pipeline_report_tag
        self.clear_task_item(task_len)
        self.current_task.task_status = TaskStatus.FAIL
        return True

    def fail_pipeline(self, *args):
        """
        若无正在运行的任务，清理pipeline的调度信息，打包运行结果，汇报结果
        若有正在运行的任务，终止此次处理，通报错误
        :param args:
        :return:
        """
        # set pipeline status is fail
        self.set_task_item(0, "task_status", TaskStatus.FAIL)

        # set task errors
        task_errors = self.get_task_item(0, hash_key="task_errors")
        if isinstance(task_errors, list) is False:
            task_errors = []
        task_errors.append(join_decode(args))
        self.set_task_item(0, hash_key="task_errors", hash_value=task_errors)
        # 若error_continue为False尽最大可能删除正在运行的任务
        error_continue = self.get_task_item(0, hash_key="error_continue")
        if error_continue is not True:  # 如果任务设置error_continue不为True Pipeline有失败时，尝试删除已放入队列的任务
            self.try_remove_running_task()
        self.try_finish_pipeline()
        self.set_current_task_error(*args)

    def package_task_item(self, task_len=None):
        if task_len is None:
            task_len = self.get_task_item(0, hash_key="task_len")
        if task_len is None:
            return
        pipeline_task = dict(task_list=[])
        pipeline_task.update(self.get_task_item(0))
        for index in range(task_len):
            pipeline_task["task_list"].append(self.get_task_item(index + 1))

        task_errors = self.get_task_item(0, hash_key="task_errors")
        if isinstance(task_errors, list) is True:
            self.current_task.add_error_msg(*task_errors)
        self.current_task.task_name = self.get_task_item(0, hash_key="task_name")
        self.current_task.start_time = pipeline_task["start_time"]
        self.current_task.sub_task_detail = pipeline_task["task_list"]

    def clear_task_item(self, task_len):
        self.task_log("Start Clear Pipeline Task Item, Task Len Is ", task_len)
        for index in range(task_len + 1):
            self.del_task_item(index)

    def analysis_ref(self, ref_str, current_index, task_len):
        split_d = self.split_ref(ref_str)
        if split_d is None:
            return False, "Input Not Standard Ref Result Format %s" % ref_str
        ref_index = split_d["index"]
        ref_key = split_d["key"]
        if isinstance(current_index, int):
            if ref_index == current_index + 1:
                return False, "Input Can Not Ref Self %s" % ref_str
        if ref_index > task_len:
            return False, "Input Ref Task %s Out Of Index %s" % (ref_index, ref_str)
        if self.get_task_item(ref_index, "task_status") != TaskStatus.SUCCESS and ref_index > 0:
            return True, None
        # 判断 是获得 input 还是 output
        if ref_index == 0:
            if self.has_task_item(ref_index, hash_key="input_%s" % ref_key) is False:
                return False, "Input Ref %s Not In Task %s Input. %s" % (ref_key, ref_index, ref_str)
            ref_output = self.get_task_item(ref_index, hash_key="input_%s" % ref_key)
        else:
            if self.has_task_item(ref_index, hash_key="output_%s" % ref_key) is False:
                work_tag = self.get_task_item(ref_index, hash_key="work_tag")
                return False, "Input Ref %s Not In Task %s %s Output. [%s]" % (ref_key, ref_index, work_tag, ref_str)
            ref_output = self.get_task_item(ref_index, hash_key="output_%s" % ref_key)
        if is_string(ref_output) is True and ref_output.startswith("&") is True:
            return False, "Ref Output Value Can Not Start With &. [%s]" % ref_output
        if isinstance(ref_output, list) is True:
            for item in ref_output:
                if is_string(item) is True and item.startswith("&") is True:
                    work_tag = self.get_task_item(ref_index, hash_key="work_tag")
                    msg = "Ref task %s %s output %s value is list, each item can not start with &, but exist item " \
                          "value is %s" % (ref_index, work_tag, ref_key, item)
                    return False, msg
        return True, dict(ref_output=ref_output, ref_index=ref_index, ref_key=ref_key)

    def convert_repeat(self, task_item, index):
        input_list_keys = []
        other_keys = ["work_tag"]
        task_output = task_item.get("task_output", dict())
        for item_key in task_item.keys():
            if item_key.startswith("output_"):
                task_output[item_key[7:]] = task_item[item_key]
                continue
            if item_key.startswith("input_") is False:
                continue
            if isinstance(task_item[item_key], list):
                input_list_keys.append(item_key)
            else:
                other_keys.append(item_key)
        task_item["task_output"] = task_output
        if "task_list" in task_item:
            other_keys.append("task_list")
        if task_item["task_type"].endswith("pipeline"):
            other_keys.append("task_output")
        if "repeat_freq" in task_item:
            repeat_freq = task_item["repeat_freq"]
        elif len(input_list_keys) <= 0:
            repeat_freq = 1
        else:
            repeat_freq = max(map(lambda x: len(task_item[x]), input_list_keys))
        for list_key in input_list_keys:
            k_l = len(task_item[list_key])
            if repeat_freq % k_l != 0:
                self.set_task_item(index + 1, "task_status", TaskStatus.INVALID)
                self.fail_pipeline("Task ", index + 1, " list input length different")
            task_item[list_key] *= repeat_freq / k_l
        pipeline_task = dict(task_list=[], task_output=dict(), task_type="pipeline", work_tag=self.work_tag)
        output_ref_def = dict()
        for output_key in task_output:
            output_value = task_output[output_key]
            if task_item["task_type"].endswith("pipeline"):
                output_ref_def[output_key] = output_key
                pipeline_task["task_output"][output_key] = []
                continue
            if is_string(output_value) is False:
                continue
            if output_value.startswith("&"):
                ov_f = re.findall("^\\d*&*(\\w+)$", output_value[1:])
                if len(ov_f) != 1:
                    continue
                output_ref_def[output_key] = ov_f[0]
                pipeline_task["task_output"][output_key] = []
            else:
                pass
        for r_index in range(repeat_freq):
            sub_task_item = dict()
            for list_key in input_list_keys:
                sub_task_item[list_key] = task_item[list_key][r_index]
            for other_key in other_keys:
                sub_task_item[other_key] = task_item[other_key]
            for o_key in output_ref_def:
                pipeline_task["task_output"][o_key].append("&%s%s" % (r_index + 1, output_ref_def[o_key]))
            pipeline_task["task_list"].append(sub_task_item)
        return pipeline_task

    def handle_task(self, key, params):
        if self.current_task.is_report_task is False:
            self.task_log("Start Format Pipeline")
            self.format_pipeline(key, params)

        # 获得task_len
        task_len = self.get_task_item(0, "task_len")
        self.task_log("Task Len Is ", task_len)
        for index in range(task_len):
            self.task_log("Start Set Input For Task ", index + 1)
            task_item = self.get_task_item(index + 1)
            if "task_status" in task_item:
                self.task_log("Task ", index + 1, " Not Need Set Input, Status Is ", task_item["task_status"])
                continue
            for item_key in task_item.keys():
                if item_key.startswith("input_") is False:
                    continue
                inp = task_item[item_key]
                if is_string(inp) is True and inp.startswith("&"):
                    self.task_log("Task ", index + 1, " Handle Input ", item_key)
                    ref_r, ref_info = self.analysis_ref(inp[1:], index, task_len)
                    if ref_r is False:
                        self.fail_pipeline("Task ", index + 1, " ", ref_info)
                    if ref_info is None:
                        continue
                    ref_output = ref_info["ref_output"]
                    ref_index = ref_info["ref_index"]
                    ref_key = ref_info["ref_key"]
                    self.task_log("Task ", index + 1, " Input ", item_key, " Ref Task", ref_index, " ", ref_key, " ",
                                  ref_output)
                    self.set_task_item(index + 1, item_key, ref_output)
                elif isinstance(inp, list):
                    for sub_i in range(len(inp)):
                        sub_inp = inp[sub_i]
                        if is_string(sub_inp) is False or sub_inp.startswith("&") is False:
                            continue
                        ref_r, ref_info = self.analysis_ref(sub_inp[1:], index, task_len)
                        if ref_r is False:
                            self.fail_pipeline("Task ", index + 1, " ", ref_info)
                        if ref_info is None:
                            continue
                        ref_output = ref_info["ref_output"]
                        self.set_task_item(index + 1, "%s_%s" % (item_key, sub_i), ref_output)

        running_count = 0
        success_count = 0
        for index in range(task_len):
            task_item = self.get_task_item(index + 1)
            if "task_status" in task_item:
                if task_item["task_status"] == TaskStatus.SUCCESS:
                    success_count += 1
                    continue
                elif task_item["task_status"] == TaskStatus.RUNNING:
                    running_count += 1
                    continue
                else:
                    continue
            is_ready = True
            input_keys = []
            for item_key in task_item.keys():
                if item_key.startswith("input_") is False:
                    continue
                input_keys.append(item_key)
                inp = task_item[item_key]
                if is_string(inp) is True and inp.startswith("&"):
                    is_ready = False
                    break
                elif isinstance(inp, list):
                    for sub_i in range(len(inp)):
                        sub_inp = inp[sub_i]
                        if is_string(sub_inp) and sub_inp.startswith("&"):
                            ref_output = self.get_task_item(index + 1, "%s_%s" % (item_key, sub_i))
                            if ref_output is None:
                                is_ready = False
                                break
                            else:
                                inp[sub_i] = ref_output
                    if is_ready is False:
                        break
            if is_ready is True:
                l = self.set_task_item(index + 1, "task_status", TaskStatus.RUNNING, nx=True)
                if l == 1:
                    if task_item["task_type"].startswith("repeat-"):
                        sub_task_params = self.convert_repeat(task_item, index)
                    else:
                        sub_task_params = task_item
                        for input_key in input_keys:
                            sub_task_params[input_key[6:]] = sub_task_params[input_key]
                    if self.current_task.task_sub_key is None:
                        sub_key = index + 1
                    else:
                        sub_key = "%s_%s" % (self.current_task.task_sub_key, index + 1)
                    self.task_log("Push Task ", index + 1, " ", task_item["work_tag"], " Run")
                    self.set_task_item(index + 1, "begin_time", time())
                    self.push_task(key, sub_task_params, sub_key=sub_key, work_tag=sub_task_params["work_tag"],
                                   report_tag=self.work_tag)
                running_count += 1
        if success_count == task_len:
            self.task_log("Task All Success")
            self.completed_pipeline()
        elif running_count == 0:
            task_status = self.get_task_item(0, "task_status")
            if task_status is None:
                self.task_log("Pipeline Has Endless Loop Waiting")
                self.fail_pipeline("Pipeline Has Endless Loop Waiting")
            return self.try_finish_pipeline()



#! /usr/bin/env python
# coding: utf-8

from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib
import thread
import ConfigParser
from datetime import datetime
from My_PC import pc_info

__author__ = 'ZhouHeng'


class MyEmailManager(object):
    def __init__(self, conf_dir):
        self.email_server = "ym.163.com"
        self.sender = "晶云平台"
        conf_path = conf_dir + "email_app.conf"
        self._int_app(conf_path)

    def _int_app(self, conf_path):
        config = ConfigParser.ConfigParser()
        config.read(conf_path)
        section = "admin"
        try:
            self.m_user = config.get(section, "email")
            self.m_password = config.get(section, "password")
        except ConfigParser.Error:
            self.m_user = ""
            self.m_password = ""

    def encoded(self, s, encoding):
        return s.encode(encoding) if isinstance(s, unicode) else s

    def send_mail(self, to, sub, content):
        try:
            encoding = 'utf-8'
            SMTP = smtplib.SMTP
            smtp = SMTP("smtp.%s" % self.email_server, 25)
            # smtp.set_debuglevel(True)
            user = self.m_user
            smtp.starttls()
            smtp.login(user, self.m_password)
            user = self.encoded(user, encoding)
            user = '{nick_name} <{user}>'.format(nick_name=Header(self.sender, encoding), user=user)
            msg = MIMEMultipart('alternative')
            msg['From'] = user
            msg['To'] = self.encoded(to, encoding)
            msg['Subject'] = Header(self.encoded(sub, encoding), encoding)
            msg.attach(MIMEText(self.encoded(content, encoding), "html", encoding))
            smtp.sendmail(user, to, msg.as_string())
            smtp.quit()
            return True
        except Exception, e:
            error_message = "MyEmailManager send_mail error %s" % str(e.args)
            print(error_message)
            return False

    def send_mail_thread(self, to, sub, content):
        return thread.start_new_thread(self.send_mail, (to, sub, content))



#! /usr/bin/env python
# coding: utf-8

import os
import re
import sys
import json
import uuid
import types
import subprocess
from time import time, sleep
import threading
import logging
import traceback
from JYTools import StringTool
from ._exception import TaskErrorException, InvalidTaskException, WorkerTaskParamsKeyNotFound
from ._Task import TaskStatus, WorkerTask, WorkerTaskParams
from ._config import WorkerConfig, WorkerLogConfig

__author__ = 'meisanggou'


class _WorkerLog(WorkerLogConfig):
    def worker_log(self, *args, **kwargs):
        pass

    def task_log(self, *args, **kwargs):
        pass

    """
    add in 0.7.5
    """
    def task_debug_log(self, *args, **kwargs):
        kwargs.update(level="DEBUG")
        self.task_log(*args, **kwargs)


class Worker(WorkerConfig, _WorkerLog):

    """
        expect_params_type
        add in version 0.6.9
    """
    expect_params_type = None

    def __init__(self, log_dir=None, work_tag=None, **kwargs):
        WorkerConfig.__init__(self, work_tag=work_tag, **kwargs)
        _WorkerLog.__init__(self, log_dir=log_dir, **kwargs)

        if StringTool.is_string(self.work_tag) is False:
            class_name = self.__class__.__name__
            msg = "Need String work_tag. Please Set {0}.DEFAULT_WORK_TAG=yourWorkTag Or {0}(work_tag=yourWorkTag)"
            raise TypeError(msg.format(class_name))
        self._id = uuid.uuid4().hex  # add in 0.9.11
        self._msg_manager = None
        self.is_running = False  # 表示worker是否已经开始运行，并不断接收任务,一旦运行起来，不可再进入test模式即调用test方法
        self._debug = False
        self.before_handler_funcs = []
        self.after_handler_funcs = []
        self.init_log_dir()
        self._handle_task_func = self.handle_task
        self.num_success_job = 0  # add in 0.8.1
        self.num_fail_job = 0  # add in 0.8.1
        self.num_wrongful_job = 0  # add in 0.8.1
        self.num_invalid_job = 0  # add in 0.8.1
        self.num_null_job = 0  # add in 0.8.1
        if "worker_index" in kwargs:
            self.worker_index = kwargs["worker_index"]
        if "redirect_stdout" in kwargs:
            self.redirect_stdout = kwargs["redirect_stdout"]
        self.heartbeat_key = self.heartbeat_prefix_key + "_" + self.work_tag
        self.queue_key = self.queue_prefix_key + "_" + self.work_tag
        self.clock_key = self.clock_prefix_key + "_" + self.work_tag + "_" + self._id
        self.current_task = WorkerTask()

    """
    add in 0.4.0
    """
    def init_log_dir(self):
        if self.log_dir is not None:
            exclusive_log_dir = os.path.join(self.log_dir, self.work_tag.lower())
            if os.path.isdir(exclusive_log_dir):
                self.log_dir = exclusive_log_dir
            else:
                try:
                    os.mkdir(exclusive_log_dir)
                    self.log_dir = exclusive_log_dir
                except OSError:
                    pass

    """
    property
    add in 0.6.9
    """

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, v):
        if self.is_running is True:
            return
        if not isinstance(v, bool):
            raise TypeError("need bool value for debug")
        self._debug = v
        if self.debug is True:
            self.redirect_stdout = False

    @property
    def num_total_job(self):
        r_job = self.num_worked_job
        t_job = r_job + self.num_wrongful_job + self.num_null_job
        return t_job

    @property
    def num_worked_job(self):
        return self.num_success_job + self.num_fail_job + self.num_invalid_job

    def has_heartbeat(self):
        return True

    def write(self, *args, **kwargs):
        self.task_log(*args, **kwargs)

    def push_task(self, key, params, work_tag=None, sub_key=None, is_report=False):
        pass

    @staticmethod
    def _subprocess_timeout_thread(p, timeout):
        """
        add in version 0.7.7
        :param p:
        :param timeout:
        :return:
        """
        while timeout > 0:
            r_code = p.poll()
            if r_code is not None:
                return
            timeout -= 1
            sleep(1)
        p.kill()
        return

    def execute_subprocess(self, cmd, stdout=None, stderr=None, error_continue=False, timeout=None):
        self.task_debug_log(cmd)
        if isinstance(cmd, list) is True:
            cmd = map(lambda x: str(x) if isinstance(x, int) else x, cmd)
        std_out = stdout
        std_err = stderr
        if std_out is None:
            std_out = subprocess.PIPE
        if std_err is None:
            if std_out == subprocess.PIPE:
                std_err = subprocess.STDOUT
            else:
                std_err = subprocess.PIPE
        child = subprocess.Popen(cmd, stderr=std_err, stdout=std_out)
        if isinstance(timeout, int) and timeout > 0:
            t_timeout = threading.Thread(target=self._subprocess_timeout_thread, args=(child, timeout))
            t_timeout.start()
        else:
            t_timeout = None

        if child.stdout is not None:
            std_log = child.stdout
        elif child.stderr is not None:
            std_log = child.stderr
        else:
            std_log = None
        exec_msg = ""
        while std_log:
            out_line = std_log.readline()
            if out_line is None or len(out_line) <= 0:
                break
            exec_msg += out_line
            self.task_log(out_line)
        child.wait()
        if t_timeout is not None:
            t_timeout.join()
        r_code = child.returncode
        if r_code != 0:
            if error_continue is False:
                self.set_current_task_error(cmd[0], " exit code not 0, is ", r_code)
            else:
                self.task_debug_log(cmd[0], " exit code not 0, is ", r_code, " but continue return.")
        else:
            self.task_debug_log(cmd[0], " exit code 0")
        return r_code, exec_msg

    def _execute(self):
        self.worker_log("Start Execute", self.current_task.task_key)
        self.hang_up_clock(1)
        self.current_task.start_time = time()
        standard_out = None
        try:
            for func in self.before_handler_funcs:
                func()
            if self.redirect_stdout is True:
                standard_out = sys.stdout
                sys.stdout = self
            self.current_task.task_status = TaskStatus.RUNNING
            if self.current_task.is_report_task is False:
                self._handle_task_func(self.current_task.task_key, self.current_task.task_params)
            else:
                self.handle_report_task()
            if self.current_task.task_status == TaskStatus.RUNNING:
                self.current_task.task_status = TaskStatus.SUCCESS
            if standard_out is not None:
                sys.stdout = standard_out
            for func in reversed(self.after_handler_funcs):
                func()
            self.num_success_job += 1
        except WorkerTaskParamsKeyNotFound as pk:
            self.current_task.task_status = TaskStatus.FAIL
            self.current_task.task_message = "Need Key %s, Not Found." % pk.missing_key
            self.task_log(self.current_task.task_message, level="ERROR")
            self.num_invalid_job += 1
        except TaskErrorException as te:
            self.current_task.task_status = TaskStatus.FAIL
            self.current_task.task_message = te.error_message
            self.worker_log("Task: ", te.key, "Params: ", te.params, " Error Info: ", te.error_message)
            self.task_log(te.error_message, level="ERROR")
            self.num_fail_job += 1
        except InvalidTaskException as it:
            self.current_task.task_status = TaskStatus.INVALID
            self.current_task.task_message = it.invalid_message
            self.task_log(it.invalid_message, level="WARING")
            self.worker_log("Invalid Task ", it.task_info, " Invalid Info: ", it.invalid_message)
            self.num_invalid_job += 1
        except Exception as e:
            self.current_task.task_status = TaskStatus.FAIL
            self.current_task.task_message = str(e)
            self.task_log(traceback.format_exc(), level="ERROR")
            self._execute_error(e)
            self.num_fail_job += 1
        except SystemExit as se:
            self.current_task.task_status = TaskStatus.FAIL
            self.current_task.task_message = str(se)
            self.task_log(traceback.format_exc(), level="ERROR")
            self.num_fail_job += 1
        finally:
            if standard_out is not None:
                sys.stdout = standard_out
            self.current_task.end_time = time()
            if self.current_task.is_report_task is False and self.current_task.task_report_tag is not None:
                self.task_debug_log("Start Report Task Status")
                self.push_task(self.current_task.task_key, self.current_task.to_dict(),
                               work_tag=self.current_task.task_report_tag, sub_key=self.current_task.task_sub_key,
                               is_report=True)
        use_time = self.current_task.end_time - self.current_task.start_time
        self.task_debug_log("Use ", use_time, " Seconds")
        self.worker_log("Completed Task", self.current_task.task_key)
        task_output = self.current_task.task_output
        task_status = self.current_task.task_status
        self.current_task = None
        return task_status, task_output

    def _execute_error(self, e):
        if self.handler_task_exception is not None:
            self.handler_task_exception(e)

    # 待废弃 被handle_task替代
    def handler_task(self, key, params):
        pass

    # 子类需重载的方法
    def handle_task(self, key, params):
        self.handler_task(key, params)

    # 待废弃 被handle_report_task替代
    def handler_report_task(self):
        """
            add in version 0.1.19
        """
        pass

    def handle_report_task(self):
        self.handler_report_task()

    # 子类需重载的方法
    def handler_task_exception(self, e):
        pass

    def handle_invalid_task(self, task_info, error_info):
        pass

    def hang_up_clock(self, freq=None):
        pass

    def hang_down_clock(self):
        pass

    def set_current_task_invalid(self, *args):
        """
            add in version 0.1.14
        """
        if self.current_task.task_key is not None:
            raise InvalidTaskException(self.current_task.task_key, self.current_task.task_params, self.current_task,
                                       *args)

    def set_current_task_error(self, *args):
        """
            add in version 0.1.18
        """
        if self.current_task.task_key is not None:
            raise TaskErrorException(self.current_task.task_key, self.current_task.task_params, *args)

    def set_output(self, key, value):
        self.task_debug_log("Task Out ", key, ": ", value)
        if isinstance(self.current_task, WorkerTask):
            self.current_task.task_output[key] = value

    def set_multi_output(self, **kwargs):
        for key, value in kwargs.items():
            self.set_output(key, value)

    @property
    def msg_manager(self):
        return self._msg_manager

    @msg_manager.setter
    def msg_manager(self, msg_manager):
        if msg_manager is None:
            return
        if hasattr(msg_manager, "publish_message") is False:
            return
        if isinstance(msg_manager.publish_message, types.MethodType) is False:
            return
        self._msg_manager = msg_manager

    def publish_message(self, message):
        """

        add in version 0.1.4
        """
        if self.msg_manager is None:
            return
        try:
            self.msg_manager.publish_message(message, self.work_tag)
        except Exception as e:
            logging.error(e)

    def run(self):
        pass

    def test(self, key, params=None, params_path=None, sub_key=None, report_tag=None, debug=True):
        if self.is_running is True:  # 一旦运行起来，不可再进入test模式即调用test方法
            raise RuntimeError("Can not test, current is running")
        self.debug = debug
        if params is None and params_path is not None:
            with open(params_path, "r") as rp:
                c = rp.read()
                params = json.loads(c)
        task_item = WorkerTask(task_key=key, task_sub_key=sub_key, task_report_tag=report_tag, work_tag=self.work_tag)
        if self.expect_params_type is not None:
            if not isinstance(params, self.expect_params_type):
                raise TypeError("params should", self.expect_params_type)
        if isinstance(params, dict):
            task_item.set(task_params=WorkerTaskParams(**params))
            task_item.task_params.debug_func = self.task_debug_log
        else:
            task_item.set(task_params=params)
        if StringTool.is_string(self.log_dir) is True:
            log_name = StringTool.join_encode([self.work_tag, "_", task_item.task_key, ".log"], join_str="")
            task_item.log_path = StringTool.path_join(self.log_dir, log_name)
        self.current_task = task_item
        return self._execute()

    def work(self, daemon=False):
        """
        add in version 0.1.8
        """
        if daemon is not False:
            self.debug = False
            try:
                pid = os.fork()
                if pid == 0:  # pid大于0代表是父进程 返回的是子进程的pid pid==0为子进程
                    self.run()
            except OSError:
                sys.exit(1)
        else:
            self.run()

    def close(self, exit_code=0):
        self.is_running = False
        self.hang_down_clock()
        self.worker_log("start close. exit code: %s" % exit_code)
        exit(exit_code)

"""
    ReadWorkerLog Add In Version 1.0.4
"""


class ReadWorkerLog(WorkerLogConfig):

    log_pattern = r"^\[[\s\S]+?\](\[[\s\S]*?\]|) (\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}): ([a-z]{1,10}) ([\s\S]*)"
    log_compile = re.compile(log_pattern, re.I)
    log_level = dict(DEBUG=("DEBUG", "INFO", "WARING", "WARNING", "ERROR"), INFO=("INFO", "WARING", "WARNING", "ERROR"),
                     WARNING=("WARING", "WARNING", "ERROR"), ERROR=("ERROR", ))

    def read_task_log(self, work_tag, key, sub_key=None, sub_key_prefix=None, level="INFO"):
        """

        :param work_tag:
        :param key:
        :param sub_key: 为None时查询所有有子key和无子key的日志，为空字符串时仅查询无子key的日志，为具体某个子key时查询具体子key的日志
        :param level: 默认为INFO，允许DEBUG，INFO，WARNING，ERROR。其他值认为是INFO
        :return:
        """
        name = StringTool.join([work_tag, "_", key, ".log"], "")
        log_path = StringTool.path_join(self.log_dir, work_tag.lower(), name)
        if os.path.exists(log_path) is False:
            log_path = StringTool.path_join(self.log_dir, name)
            if os.path.exists(log_path) is False:
                return False, None
        # 处理参数
        if sub_key is not None:
            sub_key = StringTool.encode(sub_key)
        if sub_key_prefix is not None:
            sub_key_prefix = StringTool.encode(sub_key_prefix)
        if StringTool.is_string(level) is False:
            level = "INFO"
        level = level.upper()
        if level not in self.log_level:
            level = "INFO"
        allow_levels = self.log_level[level]
        logs_list = []
        last_save = False
        with open(log_path, "r") as rl:
            all_lines = rl.readlines()
            for line in all_lines:
                rl = self.log_compile.match(line)
                if rl is not None:
                    line_sub_key = rl.groups()[0]
                    log_time = rl.groups()[1]
                    if len(line_sub_key) >= 2:
                        line_sub_key = line_sub_key[1:-1]
                    line_level = rl.groups()[2]
                    log_msg = rl.groups()[3]
                    if sub_key is not None and sub_key != line_sub_key:
                        last_save = False
                        continue
                    if sub_key_prefix is not None and line_sub_key.startswith(sub_key_prefix) is False:
                        last_save = False
                        continue
                    if line_level not in allow_levels:
                        last_save = False
                        continue
                    last_save = True
                    logs_list.append(map(StringTool.decode, [line_sub_key, log_time, line_level, log_msg]))
                elif last_save is True:
                    logs_list[-1][3] = StringTool.join_decode([logs_list[-1][3], line])
        return True, logs_list



#! /usr/bin/env python
# coding: utf-8

import types
from ._exception import WorkerTaskParamsKeyNotFound, WorkerTaskParamsValueTypeError

__author__ = '鹛桑够'


class TaskStatus(object):
    """
        add in version 0.1.19
    """
    NONE = "None"
    SUCCESS = "Success"
    FAIL = "Fail"
    ERROR = "Fail"
    INVALID = "Invalid"
    RUNNING = "Running"


class WorkerTaskParams(dict):
    """
        add in version 0.5.0
    """

    def __init__(self, seq=None, **kwargs):
        if seq is not None:
            super(WorkerTaskParams, self).__init__(seq, **kwargs)
        else:
            super(WorkerTaskParams, self).__init__(**kwargs)
        self.debug_func = None

    def get(self, k, d=None):
        v = dict.get(self, k, d)
        if isinstance(self.debug_func, types.MethodType) is True:
            self.debug_func(k, v)
        return v

    def getint(self, k, d=None):
        """
        add in version 0.7.10
        """
        v = self.get(k, d)
        if isinstance(v, int) is False:
            raise WorkerTaskParamsValueTypeError(k, v, int)
        return v

    def getboolean(self, k, d=None):
        """
        add in version 0.7.10
        """
        v = self.get(k, d)
        if isinstance(v, bool) is False:
            raise WorkerTaskParamsValueTypeError(k, v, bool)
        return v

    def getlist(self, k, d=None):
        """
        add in version 0.7.11
        """
        v = self.get(k, d)
        if isinstance(v, list) is False:
            raise WorkerTaskParamsValueTypeError(k, v, list)
        return v

    def __getitem__(self, item):
        if item not in self:
            raise WorkerTaskParamsKeyNotFound(item)
        v = dict.__getitem__(self, item)
        if isinstance(self.debug_func, types.MethodType) is True:
            self.debug_func(item, v)
        return v


class WorkerTask(object):
    """
        add in version 0.1.19
        task_name add in version 0.2.6
    """
    __slots__ = ("task_key", "task_name", "task_sub_key", "task_info", "task_params", "task_status", "task_report_tag",
                 "is_report_task", "task_output", "task_message", "task_errors", "work_tag", "start_time", "end_time",
                 "sub_task_detail", "log_path")

    def __init__(self, **kwargs):
        self.task_key = None
        self.task_name = None
        self.task_sub_key = None
        self.task_info = None
        self.task_params = None
        self.task_status = TaskStatus.NONE
        self.task_report_tag = None  # 任务结束后汇报的的work_tag
        self.is_report_task = False
        self.task_output = dict()
        self.task_message = None  # 保存任务的执行结果的综述
        self.task_errors = []  # 保存多条错误记录
        self.work_tag = None
        self.start_time = None  # 任务真正执行的开始时间
        self.end_time = None  # 任务真正执行结束的时间
        self.sub_task_detail = None
        self.log_path = None  # add in 1.1.8
        self.set(**kwargs)

    def set(self, **kwargs):
        allow_keys = ["task_key", "task_status", "task_name", "sub_task_detail", "task_sub_key", "task_info",
                      "task_params", "task_report_tag", "is_report_task", "work_tag", "task_message", "start_time",
                      "end_time", "task_output", "task_errors"]
        for k, v in kwargs.items():
            if k not in allow_keys:
                continue
            self.__setattr__(k, v)

    def to_dict(self):
        d = dict()
        d["task_key"] = self.task_key
        d["task_sub_key"] = self.task_sub_key
        # d["task_info"] = self.task_info
        # d["task_params"] = self.task_params
        d["task_name"] = self.task_name
        d["task_status"] = self.task_status
        d["task_output"] = self.task_output
        d["work_tag"] = self.work_tag
        d["task_message"] = self.task_message
        d["task_errors"] = self.task_errors
        d["start_time"] = self.start_time
        d["end_time"] = self.end_time
        d["sub_task_detail"] = self.sub_task_detail
        return d

    def __getitem__(self, item):
        return self.to_dict()[item]

    def __contains__(self, item):
        return item in self.to_dict()

    def __setitem__(self, key, value):
        kwargs = {key: value}
        self.set(**kwargs)

    def __eq__(self, other):
        if isinstance(other, WorkerTask) is False:
            return False
        if other.task_key != self.task_key:
            return False
        if other.task_sub_key != self.task_sub_key:
            return False
        return True

    def add_error_msg(self, *args):
        for arg in args:
            self.task_errors.append(arg)


if __name__ == "__main__":
    a = dict({"a": "b"}, c="c")
    print(a)
    wp = WorkerTaskParams(dict(a=1, b=2), c=5)
    print(wp)
    # for key in wp:
    #     print(key)
    # print wp.keys()
    # print(wp["a"])
    # print(wp["c"])



#! /usr/bin/env python
# coding: utf-8

import thread
import requests
from json import dumps as json_dumps

__author__ = 'ZhouHeng'


class RequestsManager(object):
    def __init__(self, conn_error_code=-1, http_error_code=1):
        self._req = requests.session()
        self.conn_error_code = conn_error_code
        self.http_error_code = http_error_code
        self.verify_http_code = True

    @property
    def auth(self):
        return self._req.auth

    @auth.setter
    def auth(self, v):
        self._req.auth = v

    @property
    def headers(self):
        return self._req.headers

    @headers.setter
    def headers(self, v):
        self._req.headers = v

    def options(self, url, **kwargs):
        return self._req.options(url, **kwargs)

    def head(self, url, **kwargs):
        return self._req.head(url, **kwargs)

    def get(self, url, **kwargs):
        return self.request("get", url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs):
        return self.request("post", url, data=data, json=json, **kwargs)

    def put(self, url, data=None, **kwargs):
        return self.request("put", url, data=data, **kwargs)

    def delete(self, url, **kwargs):
        return self.request("delete", url, **kwargs)

    def close(self):
        self._req.close()

    def request(self, method, url, **kwargs):
        as_thread = kwargs.pop("as_thread", False)
        if as_thread is True:
            return thread.start_new_thread(self.request, (method, url), kwargs)
        if "allow_redirects" not in kwargs:
            kwargs["allow_redirects"] = True
        body = kwargs.pop("body", None)
        if body is not None:
            if method == "GET":
                kwargs["params"] = body
            elif method == "POST" or method == "GET" or method == "DELETE":
                kwargs["json"] = body
        try:
            resp = self._req.request(method, url, **kwargs)
        except requests.ConnectionError as ce:
            if hasattr(ce.message, "reason") is True:
                msg = ce.message.reason
            else:
                msg = ce.message
            raise JYRequestsException(self.conn_error_code, url, message=msg, **kwargs)
        if self.verify_http_code is True and resp.status_code != 200:
            raise JYRequestsException(self.http_error_code, url, http_code=resp.status_code, **kwargs)
        return resp


class JYRequestsException(Exception):
    def __init__(self, error_type, url, **kwargs):
        self.error_type = error_type
        self.url = url
        if "http_code" in kwargs:
            self.http_code = kwargs["http_code"]
        else:
            self.http_code = 0
        if "message" in kwargs:
            self.message = str(kwargs["message"])
        else:
            self.message = ""
        self.json = None
        self.data = None
        if "json" in kwargs:
            self.json = kwargs["json"]
        if "data" in kwargs:
            self.data = kwargs["data"]

    def __str__(self):
        exp_msg = {"url": self.url, "error": self.message, "http_code": self.http_code}
        if self.data is not None:
            exp_msg["data"] = self.data
        if self.json is not None:
            exp_msg["json"] = self.json
        return json_dumps(exp_msg)



#! /usr/bin/env python
# coding: utf-8
import sys
from JYTools.JYWorker._config import RedisWorkerConfig, WorkerConfig
if sys.version_info[0] == 3:
    raw_input = input

"""
add in version 0.8.7
"""
redis_host_input = "please input the redis_host, need the format like 192.168.0.1, default localhost"
redis_password_input = "please input the redis_password, need string redis_password, default null"
redis_port_input = "please input the redis_port, need number redis_port < 65535, default 6379"
redis_db_input = "please input the redis_db, need number redis_db <= 16, default 13"
select_key_input = "if you want to change the value, input the number ahead, or you can input w to finish:"
heartbeat_prefix_key_input = "please input the heartbeat_prefix_key, need string heartbeat_prefix_key, " \
                             "default worker_heartbeat"
queue_prefix_key_input = "please input the queue_prefix_key, need string queue_prefix_key, default task_queue"
judge_prefix_input = "either the queue_prefix_key or the heartbeat_prefix_key cannot contain the other, input again"
pop_time_out_input = "please input the pop_time_out, need number pop_time_out, default 60"
file_path_input = "please input the config file path, need string config file path ,default redis.conf:"
previous_set = "Previously set:"


def create_config():
    file_path = raw_input("%s" % file_path_input)
    if not file_path.strip():
        file_path = 'redis.conf'
    redis_host = raw_input("%s:" % redis_host_input)
    num = redis_host.split('.')
    len_num = len(num)
    if not redis_host.strip() or redis_host == "localhost":
        redis_host = 'localhost'
    else:
        while len_num != 4:
            redis_host = raw_input("%s:" % redis_host_input)
            if not redis_host.strip() or redis_host == "localhost":
                redis_host = 'localhost'
                break
            else:
                num = redis_host.split('.')
                len_num = len(num)
        while len_num == 4 and not (num[0].isdigit() and num[1].isdigit() and num[2].isdigit() and num[3].isdigit()):
            redis_host = raw_input("%s:" % redis_host_input)
            if not redis_host.strip() or redis_host == "localhost":
                redis_host = 'localhost'
            else:
                num = redis_host.split('.')
                len_num = len(num)
                while len_num != 4:
                    redis_host = raw_input("%s:" % redis_host_input)
                    if not redis_host.strip() or redis_host == "localhost":
                        redis_host = 'localhost'
                        break
                    else:
                        num = redis_host.split('.')
                        len_num = len(num)
    redis_password = raw_input("%s:" % redis_password_input)
    if not redis_password.strip():
        redis_password = ''
    redis_port = raw_input("%s:" % redis_port_input)
    if not redis_port.strip():
        redis_port = '6379'
    else:
        while not redis_port.isdigit() or int(redis_port) >= 65535 or int(redis_port) <= 0:
            redis_port = raw_input("%s:" % redis_port_input)
            if not redis_port.strip():
                redis_port = '6379'
    redis_db = raw_input("%s:" % redis_db_input)
    if not redis_db.strip():
        redis_db = '13'
    else:
        while not redis_db.isdigit() or int(redis_db) > 16 or int(redis_db) <= 0:
            redis_db = raw_input("%s:" % redis_db_input)
            if not redis_db.strip():
                redis_db = '13'
    heartbeat_prefix_key = raw_input("%s:" % heartbeat_prefix_key_input)
    if not heartbeat_prefix_key.strip():
        heartbeat_prefix_key = 'worker_heartbeat'
    queue_prefix_key = raw_input("%s:" % queue_prefix_key_input)
    if not queue_prefix_key.strip():
        queue_prefix_key = 'task_queue'
    while queue_prefix_key in heartbeat_prefix_key or heartbeat_prefix_key in queue_prefix_key:
        queue_prefix_key = raw_input("%s:" % judge_prefix_input)
    pop_time_out = raw_input("%s:" % pop_time_out_input)
    if not pop_time_out.strip():
        pop_time_out = '60'
    else:
        while not pop_time_out.isdigit():
            pop_time_out = raw_input("%s:" % pop_time_out_input)
            if not pop_time_out.strip():
                pop_time_out = '60'
    select = True
    while select:
        print ("1 redis_host: %s\n2 redis_password: %s\n3 redis_port: %s\n4 redis_db:%s\n"
               "5 heartbeat_prefix_key: %s\n6 queue_prefix_key: %s\n7 pop_time_out: %s" %
               (redis_host, redis_password, redis_port, redis_db, heartbeat_prefix_key, queue_prefix_key, pop_time_out))
        select_key = raw_input("%s\n" % select_key_input)
        if select_key == '1':
            redis_host_last = redis_host
            redis_host = raw_input("%s, %s %s):" % (redis_host_input, previous_set, redis_host_last))
            num = redis_host.split('.')
            len_num = len(num)
            if not redis_host.strip():
                redis_host = redis_host_last
            else:
                while len_num != 4:
                    redis_host = raw_input("%s, %s %s):" % (redis_host_input, previous_set, redis_host_last))
                    if not redis_host.strip():
                        redis_host = redis_host_last
                        break
                    else:
                        num = redis_host.split('.')
                        len_num = len(num)
                while len_num == 4 and not \
                        (num[0].isdigit() and num[1].isdigit() and num[2].isdigit() and num[3].isdigit()):
                    redis_host = raw_input("%s, %s %s):" % (redis_host_input, previous_set, redis_host_last))
                    if not redis_host.strip():
                        redis_host = redis_host_last
                        break
                    else:
                        num = redis_host.split('.')
                        len_num = len(num)
                        while len_num != 4:
                            redis_host = raw_input("%s, %s %s):" % (redis_host_input, previous_set, redis_host_last))
                            if not redis_host.strip():
                                redis_host = redis_host_last
                                break
                            else:
                                num = redis_host.split('.')
                                len_num = len(num)
        if select_key == '2':
            redis_password_last = redis_password
            redis_password = raw_input("%s, %s %s:" % (redis_password_input, previous_set, redis_password_last))
            if not redis_password.strip():
                redis_password = redis_password_last
        if select_key == '3':
            redis_port_last = redis_port
            redis_port = raw_input("%s, %s %s:" % (redis_port_input, previous_set, redis_port_last))
            if not redis_port.strip():
                redis_port = redis_port_last
            else:
                while not redis_port.isdigit() or int(redis_port) >= 65535 or int(redis_port) <= 0:
                    redis_port = raw_input("%s, %s %s:" % (redis_port_input, previous_set, redis_port_last))
                    if not redis_port.strip():
                        redis_port = redis_port_last
        if select_key == '4':
            redis_db_last = redis_db
            redis_db = raw_input("%s, %s %s):" % (redis_db_input, previous_set, redis_db_last))
            if not redis_db.strip():
                redis_db = '13'
            else:
                while not redis_db.isdigit() or int(redis_db) > 16 or int(redis_db) <= 0:
                    redis_db = raw_input("%s, %s %s):" % (redis_db_input, previous_set, redis_db_last))
                    if not redis_db.strip():
                        redis_db = redis_db_last
        if select_key == '5':
            heartbeat_prefix_key_last = heartbeat_prefix_key
            heartbeat_prefix_key = raw_input("%s, %s %s):" % (heartbeat_prefix_key_input, previous_set,
                                                              heartbeat_prefix_key_last))
            if not heartbeat_prefix_key.strip():
                heartbeat_prefix_key = heartbeat_prefix_key_last
            while queue_prefix_key in heartbeat_prefix_key or heartbeat_prefix_key in queue_prefix_key:
                heartbeat_prefix_key = raw_input("%s, %s %s):" % (judge_prefix_input, previous_set,
                                                                  heartbeat_prefix_key_last))
        if select_key == '6':
            queue_prefix_key_last = queue_prefix_key
            queue_prefix_key = raw_input("%s, %s %s):" % (queue_prefix_key_input, previous_set, queue_prefix_key_last))
            if not queue_prefix_key.strip():
                queue_prefix_key = queue_prefix_key_last
            while queue_prefix_key in heartbeat_prefix_key or heartbeat_prefix_key in queue_prefix_key:
                queue_prefix_key = raw_input("%s:" % judge_prefix_input)
        if select_key == '7':
            pop_time_out_last = pop_time_out
            pop_time_out = raw_input("%s, %s %s):" % (pop_time_out_input, previous_set, pop_time_out_last))
            if not pop_time_out.strip():
                pop_time_out = pop_time_out_last
            else:
                while not pop_time_out.isdigit():
                    pop_time_out = raw_input("%s, %s %s):" % (pop_time_out_input, previous_set, pop_time_out_last))
                    if not pop_time_out.strip():
                        pop_time_out = pop_time_out_last
        if not select_key.strip() or select_key == 'w':
            select = False
    config = dict()
    config["file_path"] = file_path
    config["redis_host"] = redis_host
    config["redis_password"] = redis_password
    config["redis_port"] = redis_port
    config["redis_db"] = redis_db
    config["heartbeat_prefix_key"] = heartbeat_prefix_key
    config["queue_prefix_key"] = queue_prefix_key
    config["pop_time_out"] = pop_time_out
    print ("the config file path: %s\n1 redis_host: %s\n2 redis_password: %s\n3 redis_port: %s\n"
           "4 redis_db:%s\n5 heartbeat_prefix_key:%s\n6 queue_prefix_key:%s\n7 pop_time_out:%s\n" %
           (file_path, redis_host, redis_password, redis_port, redis_db, heartbeat_prefix_key, queue_prefix_key,
            pop_time_out))
    return config

if __name__ == "__main__":
    config_result = create_config()
    RedisWorkerConfig.write_config(config_result["file_path"], config_result["redis_host"],
                                   config_result["redis_password"], config_result["redis_port"],
                                   config_result["redis_db"], section_name="Redis", append=False)
    WorkerConfig.write_work_config(config_result["file_path"], config_result["heartbeat_prefix_key"],
                                   config_result["queue_prefix_key"], config_result["pop_time_out"],
                                   section_name="Worker", append=True)




#! /usr/bin/env python
# coding: utf-8

from JYTools import StringTool

__author__ = 'meisanggou'

"""
add in version 0.1.18
"""


class InvalidTaskException(Exception):
    def __init__(self, key=None, params=None, task_info=None, *args):
        self.key = key
        self.params = params
        self.task_info = task_info
        self.invalid_message = StringTool.join(args, "")


class TaskErrorException(Exception):
    def __init__(self, key, params, *args):
        self.key = key
        self.params = params
        self.error_message = StringTool.join(args, "")


class InvalidTaskKey(Exception):

    def __str__(self):
        return "Task Key Length Must Be Greater Than 0"


class InvalidWorkerTag(Exception):

    def __str__(self):
        return "Worker Tag Must Be String And Length Greater Than 0"


class WorkerTaskParamsKeyNotFound(Exception):

    def __init__(self, key):
        self.missing_key = key

    def __str__(self):
        return "Not Found Key %s" % self.missing_key


class WorkerTaskParamsValueTypeError(Exception):

    def __init__(self, k, v, t):
        self.key = k
        self.value = v
        self.except_type = t
